
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

#ifndef PBRT_OCTREESS_H
#define PBRT_OCTREESS_H
// octree.h*
#include "pbrt.h"
#include "geometry.h"
#include <vector>

//angle of tolerance
#define EPSILONSS (M_PI / 4.f)

//bah! globals!!
//extern float G_Zv;
//extern float G_Zr;
//extern float G_Sigmatr;
//extern float G_Fdt;
class COREDLL SSMaterialData {
public:
	SSMaterialData() {
		G_Zv = G_Zr = G_Sigmatr = G_Fdt = G_Fdr = 0.0;
	}
	float G_Zv; // = Zv;
	float G_Zr; // = Zr;
	float G_Sigmatr; // = sigmatr;
	float G_Fdt; // = Fdt;
	float G_Fdr; //= Fdr;
};
// OctreeSS Declarations
class COREDLL OctNodeSS {
public:
	OctNodeSS() {
		isLeaf = true;
		for (int i = 0; i < 8; ++i)
			children[i] = NULL;
	}
	~OctNodeSS() {
		for (int i = 0; i < 8; ++i)
			delete children[i];
	}
	/*
	OctNodeSS &operator=(const OctNodeSS &o) {
		for(int i=0; i < 8; ++i) {
			children[i] = o.children[i];
		}
		for(int i=0; i < data.size(); i++) {
			data[i] = o.data[i];
		}
		isLeaf = o.isLeaf;
		myBounds = o.myBounds;
	}*/

	OctNodeSS *children[8];
	vector<SamplePoint> data;
	bool isLeaf;
	BBox myBounds;

	Spectrum SamplePointContribution(SamplePoint sp, Point p, SSMaterialData smd) {
		float r2 = (p - sp.p).LengthSquared();
		r2 *= r2;
		float dv = sqrtf(r2 + smd.G_Zv*smd.G_Zv);
		float dr = sqrtf(r2 + smd.G_Zr*smd.G_Zr);
		float C1 = smd.G_Zr * (smd.G_Sigmatr + 1.f/dr);
		float C2 = smd.G_Zv * (smd.G_Sigmatr + 1.f/dv);
		float Mop = smd.G_Fdt * (C1*expf(-smd.G_Sigmatr*dr) + C2*expf(-smd.G_Sigmatr*dv)) / (4.f * M_PI);
		Spectrum L(Mop);
		return L;
	}

	Spectrum CalcRadiance(Point p, SSMaterialData smd) {
		Spectrum Mo(0.f);
		//check if leaf
		if(isLeaf) {
			for(int i=0; i < data.size(); i++) {
				Mo += SamplePointContribution(data[i], p, smd);
			}
			return Mo;
		}
		
		//check if point is inside the bounds
		if (myBounds.Inside(p)) {
			for(int i=0; i < 8; i++) {
				if(children[i])
					Mo += children[i]->CalcRadiance(p, smd);
			}
			return Mo;
		}
		//otherwise, check if the node is "small enough"
		assert(data.size() == 1);
		float deltaW = data[0].area / ((p - data[0].p).LengthSquared());
		if(deltaW < EPSILONSS) {
			Mo = SamplePointContribution(data[0], p, smd);
			return Mo;
		}
		for(int i=0; i < 8; i++) {
			if(children[i])
				Mo += children[i]->CalcRadiance(p, smd);
		}
		return Mo;
	}
};
class COREDLL OctreeSS : public ReferenceCounted {
public:
	// OctreeSS Public Methods
	OctreeSS() {
		maxDepth = 0;
	}
	OctreeSS(OctreeSS &o) {
		bound = o.bound;
		maxDepth = o.maxDepth;
		maxSamplesPerNode = o.maxSamplesPerNode;
	}
	OctreeSS(const BBox &b, int md = 16, int ms = 8)
		: bound(b) {
		maxDepth = md;
		maxSamplesPerNode = ms;
	}

	OctreeSS &operator=(const OctreeSS& o) {
		root = o.root;
		bound = o.bound;
		maxDepth = o.maxDepth;
		maxSamplesPerNode = o.maxSamplesPerNode;
		return *this;
	}

	void Add(const SamplePoint &dataItem, const Point &p) {//const BBox &dataBound) {
		addPrivateSS(&root, bound, dataItem, p);
	}
	void SumUpInternalNodes() {
		sumUpPrivate(&root);
	}
	OctNodeSS root;
//private:
	// OctreeSS Private Methods
	SamplePoint sumUpPrivate(OctNodeSS *node) {
		if(node->isLeaf) {
			SamplePoint sum;
			//assert(node->data.size() < maxSamplesPerNode); //may not be true b/c of depth
			for(int i=0 ; i < node->data.size(); i++) {
				sum.area += node->data[i].area;
				sum.irradiance += node->data[i].irradiance;
				sum.p += node->data[i].irradiance.y() * node->data[i].area * node->data[i].p;
			}
			return sum;
		}
		SamplePoint sum, temp;
		for(int i = 0; i < 8; i++) {
			if(node->children[i] == 0)
				continue;
			temp = sumUpPrivate(node->children[i]);
			sum.area += temp.area;
			sum.irradiance += temp.irradiance;
			sum.p += temp.irradiance.y() * temp.area * temp.p;
		}
		node->data.resize(1, sum);
		//node->data.clear();
		//node->data.push_back(sum);
		return sum;
	}
	void addPrivateSS(OctNodeSS *node, const BBox &nodeBound,
		const SamplePoint &dataItem, const Point &p, int depth = 0)  {
			node->myBounds = nodeBound;
			// Possibly add data item to current octree node
			if ((depth == maxDepth) || 
				(node->isLeaf	&& node->data.size() < maxSamplesPerNode)) {
				node->isLeaf = true;
				node->data.push_back(dataItem);
				return;
			}
			
			node->isLeaf = false;
			Point pMid = .5 * nodeBound.pMin + .5 * nodeBound.pMax;
			bool over[8];

			//check if there's too many samples in current node
			if(node->data.size() >= maxSamplesPerNode) {
				//reassign all data in current node to children
				//node->isLeaf = false;
				for(int i=0; i < node->data.size(); i++) {
					over[0] = over[1] =	over[2] =	over[3] = (p.x <= pMid.x);
					over[4] = over[5] =	over[6] =	over[7] = (p.x  > pMid.x);
					over[0] &= (p.y <= pMid.y);
					over[1] &= (p.y <= pMid.y);
					over[4] &= (p.y <= pMid.y);
					over[5] &= (p.y <= pMid.y);
					over[2] &= (p.y  > pMid.y);
					over[3] &= (p.y  > pMid.y);
					over[6] &= (p.y  > pMid.y);
					over[7] &= (p.y  > pMid.y);
					over[0] &= (p.z <= pMid.z);
					over[2] &= (p.z <= pMid.z);
					over[4] &= (p.z <= pMid.z);
					over[6] &= (p.z <= pMid.z);
					over[1] &= (p.z  > pMid.z);
					over[3] &= (p.z  > pMid.z);
					over[5] &= (p.z  > pMid.z);
					over[7] &= (p.z  > pMid.z);
					for (int child = 0; child < 8; ++child) {
						if (!over[child]) continue;
						if (!node->children[child]) //note that we should be in this part of the loop once
							node->children[child] = new OctNodeSS();
						// Compute _childBound_ for octree child _child_
						BBox childBound;
						childBound.pMin.x = (child & 4) ? pMid.x : nodeBound.pMin.x;
						childBound.pMax.x = (child & 4) ? nodeBound.pMax.x : pMid.x;
						childBound.pMin.y = (child & 2) ? pMid.y : nodeBound.pMin.y;
						childBound.pMax.y = (child & 2) ? nodeBound.pMax.y : pMid.y;
						childBound.pMin.z = (child & 1) ? pMid.z : nodeBound.pMin.z;
						childBound.pMax.z = (child & 1) ? nodeBound.pMax.z : pMid.z;
						addPrivateSS(node->children[child], childBound,
							node->data[i], p, depth+1);
					}
				}
				node->data.clear();
			}
			// Otherwise add data item to octree children
			// Determine which single children the item is in
			
			over[0] = over[1] =
						over[2] =
						over[3] = (p.x <= pMid.x);
			over[4] = over[5] =
						over[6] =
						over[7] = (p.x  > pMid.x);
			over[0] &= (p.y <= pMid.y);
			over[1] &= (p.y <= pMid.y);
			over[4] &= (p.y <= pMid.y);
			over[5] &= (p.y <= pMid.y);
			over[2] &= (p.y  > pMid.y);
			over[3] &= (p.y  > pMid.y);
			over[6] &= (p.y  > pMid.y);
			over[7] &= (p.y  > pMid.y);
			over[0] &= (p.z <= pMid.z);
			over[2] &= (p.z <= pMid.z);
			over[4] &= (p.z <= pMid.z);
			over[6] &= (p.z <= pMid.z);
			over[1] &= (p.z  > pMid.z);
			over[3] &= (p.z  > pMid.z);
			over[5] &= (p.z  > pMid.z);
			over[7] &= (p.z  > pMid.z);
			for (int child = 0; child < 8; ++child) {
				if (!over[child]) continue;
				if (!node->children[child])
					node->children[child] = new OctNodeSS();
				// Compute _childBound_ for octree child _child_
				BBox childBound;
				childBound.pMin.x = (child & 4) ? pMid.x : nodeBound.pMin.x;
				childBound.pMax.x = (child & 4) ? nodeBound.pMax.x : pMid.x;
				childBound.pMin.y = (child & 2) ? pMid.y : nodeBound.pMin.y;
				childBound.pMax.y = (child & 2) ? nodeBound.pMax.y : pMid.y;
				childBound.pMin.z = (child & 1) ? pMid.z : nodeBound.pMin.z;
				childBound.pMax.z = (child & 1) ? nodeBound.pMax.z : pMid.z;
				addPrivateSS(node->children[child], childBound,
									dataItem, p, depth+1);
			}
		}
	// OctreeSS Private Data
	int maxDepth;
	int maxSamplesPerNode;
	BBox bound;
};
// OctreeSS Method Definitions
//void OctreeSS::addPrivateSS(
//		OctNodeSS *node, const BBox &nodeBound,
//		const SamplePoint &dataItem,  const Point &p, int depth)
#endif // PBRT_OCTREE_H
