/*
 *  wavefront.cpp
 *  Mark Colbert
 *  ma434187@pegasus.cc.ucf.edu	
 *
 *  The following will parse triangulated obj files that can eaisly be exported from 
 *  lightwave, maya, or 3d studio max.  
 *
 *  TODO: The largest problem about this program is that if there are a different
 *	number of normal verticies or UV verticies than point verticies, which there usually are, the 
 *  program will just create as many verticies as there are in the index to 
 *  easily ensure that every point has the proper UV, normal, and point vertex
 *  information.  The fix is to just use a more efficient methodology of combining
 *  all the data.
 *  
 *  TODO: The other problem about this parser is that it only works with triangulated data.
 *	In other words, if you give it an object that is anything but triangles, it will
 *  only read the first 3 vertecies of a face and you will get one very odd looking
 *  model.  Thus, it would be nice to have some sort of triangulation algorithm in
 *  this program to allow any generic OBJ file to be input for the plugin.
 *
 */
 

#include "shape.h"
#include "paramset.h"
#include "dynload.h"
#include <vector>

using namespace std;

// line buffer size determines at compile time how large the input
// buffer should be for the file input lines
#define LINE_BUFFER_SIZE 1024

class Wavefront : public Shape {
	public:
		Wavefront(const char* filename, const Transform &o2w, bool reverseOrientation); //, float G, float SigmaS, float SigmaA);
		~Wavefront() {
			if (vertexIndex) delete[] vertexIndex;
			if (p) delete[] p;
			if (n) delete[] n;
			if (uvs) delete[] uvs;
		}
		
		BBox ObjectBound() const;
		BBox WorldBound() const;
		bool CanIntersect() const { return false; }
		void Refine(vector<Reference<Shape> > &refined) const;

		Point Wavefront::Sample(float u1, float u2, Normal *Ns) const;
		float Wavefront::Area() const { return surfaceArea; }
		//for subsurface scattering
		void Wavefront::GetSamples(int &nSamples, SamplePoint *&smpPts);
	private:
		void MergeIndicies(vector<Point> &points, vector<Normal> &normals, vector<float> &uvVec,
							vector<int> &vIndex, vector<int> &normalIndex, vector<int> &uvIndex);
		int ntris, nverts, npoints;
		int *vertexIndex;
		Point *p;
		Normal *n;
		float *uvs;
		//used for subsurface scattering
		float surfaceArea;
		vector<float> areaPartialSum;
		vector<float> faceArea; //surface area of each triangle
		float *vertexArea; //area associated with each vertex
		Normal *vertexNormal; //normal associated with each vertex
		/*
		//used for subsurface scattering
		float g; //mean cosine of scattering angle
		float sigmas; //scattering coef.
		float sigmasnot; //reduced scattering coef.
		float sigmaa; //absorption coef. 
		float sigmatnot; //reduced extinction coef.
		float lu; //mean-free path
		*/
};

// GETNUM just gets the next number from a line of input in an OBJ file
#ifndef GETNUM
#define GETNUM(lineBuffer, numBuffer, lindex, nindex, tval)  \
	nindex=0;\
	while ((lineBuffer[lindex] == ' ') || lineBuffer[lindex] == '/') lindex++;\
	while ((lineBuffer[lindex] != ' ') && (lineBuffer[lindex] != '/') && \
		   (lineBuffer[lindex] != '\0') && (lineBuffer[lindex] != '\n') && (lindex != LINE_BUFFER_SIZE)) { \
		numBuffer[nindex] = lineBuffer[lindex]; \
		nindex++; \
		lindex++; \
	} \
	numBuffer[nindex] = '\0'; \
	tval = atoi(numBuffer);
#endif

// constructor / parser
Wavefront::Wavefront(const char* filename, const Transform &o2w, bool reverseOrientation) : //, 
											//float G, float SigmaS, float SigmaA) :
		   Shape(o2w, reverseOrientation), p(NULL), n(NULL), uvs(NULL) {
	FILE* fin;
	fin = fopen(filename, "r");
	if (!fin) return;
	
	// temporary input buffers
	vector<Point> points;
	vector<int> verts;
	vector<int> normalIndex;
	vector<int> uvIndex;
	
	vector<Normal> file_normals;
	vector<float> file_uvvector;
		
	Point ptmp;
	Normal ntmp;
	float uv1, uv2;
	
	char lineBuffer[LINE_BUFFER_SIZE];
	char numBuffer[32];
	int lindex=0;
	int nindex=0;
	int ival, uvval, nval;
	ntris=0;
	surfaceArea = 0.f;
	//for surfaceArea calculation;
	Point tempp[3];
	float currentArea;
	
	// parse the data in
	while (fgets(lineBuffer, LINE_BUFFER_SIZE, fin)) {
		//cout << lineBuffer[0] << endl;
		switch (lineBuffer[0]) {
			case 'v':
				// case vertex information
				if (lineBuffer[1] == ' ') {
					// regular vertex point
					sscanf(&lineBuffer[2], "%f %f %f", &ptmp.x, &ptmp.y, &ptmp.z);
					points.push_back(ptmp);
				} else if (lineBuffer[1] == 't') {
					// texture coordinates
					sscanf(&lineBuffer[3], "%f %f", &uv1, &uv2);
					file_uvvector.push_back(uv1);
					file_uvvector.push_back(uv2);
				} else if (lineBuffer[1] == 'n') {
					// normal vector
					sscanf(&lineBuffer[2], "%f %f %f", &ntmp.x, &ntmp.y, &ntmp.z);
					file_normals.push_back(ntmp);
				}
				break;
			case 'f':
				// case face information
				lindex = 2;
				ntris++;
				for (int i=0; i < 3; i++) {
					
					GETNUM(lineBuffer, numBuffer, lindex, nindex, ival)
					
					// obj files go from 1..n, this just allows me to access the memory
					// directly by droping the index value to 0...(n-1)
					ival--;
					//for surfaceArea calculation
					tempp[i] = points[ival];

					verts.push_back(ival);
										
					if (lineBuffer[lindex] == '/') {
						lindex++;
						GETNUM(lineBuffer, numBuffer, lindex, nindex, uvval)
						uvIndex.push_back(uvval-1);
					}
					
					if (lineBuffer[lindex] == '/') {
						lindex++;
						GETNUM(lineBuffer, numBuffer, lindex, nindex, nval)
						normalIndex.push_back(nval-1);
					}
					lindex++;
				}
				//surfaceArea calculation, taken from pg.130 of pbrt book
				currentArea = 0.5f * Cross(tempp[1]-tempp[0], tempp[2]-tempp[0]).Length();
				surfaceArea += currentArea;
				faceArea.push_back(currentArea);
				areaPartialSum.push_back(surfaceArea);
				break;
			case 'g':
				// not really caring about faces or materials now
				// so just making life easier, I'll ignoring it
				break;
		}
	}
		
	fclose(fin);
	
	npoints = points.size(); //for subsurface
	//used for subsurface scattering
	assert(faceArea.size() == ntris);
	vertexArea = new float[npoints];
	vertexNormal = new Normal [npoints];
	for(int i=0; i < npoints; i++)
		vertexArea[i] = 0.f;
	float thirdArea;
	for(int i=0; i < ntris; i++) {
		thirdArea = faceArea[i] / 3.0f;
		vertexArea[verts[3*i]] += thirdArea;
		vertexArea[verts[3*i+1]] += thirdArea;
		vertexArea[verts[3*i+2]] += thirdArea;
		vertexNormal[verts[3*i]] = file_normals[normalIndex[3*i]]; //do i need to average them?
		vertexNormal[verts[3*i+1]] = file_normals[normalIndex[3*i+1]]; //do i need to average them?
		vertexNormal[verts[3*i+2]] = file_normals[normalIndex[3*i+2]]; //do i need to average them?
	}

	// merge everything back into one index array instead of multiple arrays
	MergeIndicies(points, file_normals, file_uvvector, verts, normalIndex, uvIndex);
	
	points.clear();
	file_normals.clear();
	file_uvvector.clear();
	verts.clear();
	normalIndex.clear();
	uvIndex.clear();
		
	if (n) std::cout << "Used normals" << std::endl;
	if (uvs) std::cout << "Used texture coords" << std::endl;
/*
	//used for subsurface scattering
	assert(faceArea.size() == ntris);
	vertexArea = new float[npoints];
	for(int i=0; i < npoints; i++)
		vertexArea[i] = 0.f;
	float thirdArea;
	for(int i=0; i < ntris; i++) {
		thirdArea = faceArea[i] / 3.0f;
		vertexArea[vertexIndex[3*i]] += thirdArea;
		vertexArea[vertexIndex[3*i+1]] += thirdArea;
		vertexArea[vertexIndex[3*i+2]] += thirdArea;
	}
*/


	/*	
	//used for subsurface scattering
	g = G;
	sigmas = SigmaS;
	sigmaa = SigmaA;
	if(g > 0.7f) {
		sigmasnot = (1.0f - g) * sigmas;
	}
	else
		sigmasnot = sigmas;
	sigmatnot = sigmasnot + sigmaa;
	lu = 1.0f / sigmatnot;
	
	//from paper "Generating Textures on Arbitrary Surfaces Using Reaction-Diffusion", Turk 1991
	//choose random samples on the polygon surface
	int nSamples = (int) (surfaceArea / (M_PI*lu*lu));
	Point *sampledPoints = new Point[nSamples];
	Normal *pointNormals = new Normal[nSamples];
	for(int i=0; i < nSamples; i++) {
		samplePoints[i] = Sample(RandomFloat(), RandomFloat(), &(pointNormals[i]));
	}
	//use repulsion!
	int k = 40;
	float r = 2 * sqrt(surfaceArea / nSamples);
	for(int i=0; i < k; i++) {
		for(int j=0; j < nSamples; j++) {
			//determine nearby points to P

			//map these nearby points onto the plane containing the polygon
			
			//compute and store the repulsive forces that the mapped points exert on P
		}

		for(int j=0; j < nSamples; j++) {	
			//compute the new position of P based on the repulsive forces
		}
	}

	//afterwards, make sure to convert points back to world (ObjectToWorld(p))
	*/
	//ss_myOctree = new OctreeSS();
}

void Wavefront::MergeIndicies(vector<Point> &points, vector<Normal> &normals, vector<float> &uvVec, vector<int> &vIndex, vector<int> &normalIndex, vector<int> &uvIndex) {

	bool useNormals = !normals.empty();
	bool useUVs = !uvVec.empty();
	
	
	if (!useNormals && !useUVs) { 
		std::cout << "Copying points" << std::endl;
		// just copy the points into the array	
		nverts = vIndex.size();
		int pSize = points.size();
		//cout << "point size is" << pSize << endl;
		p = (pSize > 0)? new Point[points.size()]: NULL;
		for (unsigned int i=0; i < points.size(); i++)
			p[i] = points[i];
		vertexIndex = new int[nverts];
		for (int i=0; i < nverts; i++)
			vertexIndex[i] = vIndex[i];
		return;
	}
	//cout << "done with points" << endl;
		
	// assumes that vertexIndex = normalIndex = uvIndex	
	nverts = vIndex.size();
	//cout << "nverts is: " << nverts << endl;
	vertexIndex = new int[nverts];
	//cout << "initialized vertexIndex" << endl;
	
	p = new Point[nverts];
	//cout << "initialized p" << endl;
	n = NULL; 
	uvs = NULL;
	if (useNormals) n = new Normal[nverts];
	if (useUVs) uvs = new float[nverts*2];
	
	for (int i=0; i < nverts; i++) {
		//cout << "in loop, i: " << i << endl;
		assert(vIndex[i] < nverts);
		p[i] = points[vIndex[i]];
		if (useNormals) {
			assert(normalIndex[i] < nverts);
			n[i] = normals[normalIndex[i]];
		}
		if (useUVs) { 
			uvs[i*2] = uvVec[uvIndex[i]*2];
			uvs[i*2+1] = uvVec[uvIndex[i]*2 + 1];
		}
		vertexIndex[i] = i;
	}	
}

BBox Wavefront::ObjectBound() const {
	BBox bobj;
	for (int i = 0; i < nverts; i++)
		bobj = Union(bobj, WorldToObject(p[i]));
	return bobj;
}

BBox Wavefront::WorldBound() const {
	BBox worldBounds;
	for (int i = 0; i < nverts; i++)
		worldBounds = Union(worldBounds, p[i]);
	return worldBounds;
}

// Refine
// generates the triangle mesh shape
void Wavefront::Refine(vector<Reference<Shape> > &refined) const {
	ParamSet paramSet;
	paramSet.AddInt("indices", vertexIndex, ntris*3);
	paramSet.AddPoint("P", p, nverts);		
	if (n) paramSet.AddNormal("N", n, nverts);
	if (uvs) paramSet.AddFloat("uv", uvs, nverts*2);
	//changed for subsurface scattering
	Reference<Shape> temp = MakeShape("trianglemesh", ObjectToWorld, reverseOrientation, paramSet);
	temp->ss_n = this->ss_n;
	temp->ss_smp = this->ss_smp;
	//temp->ss_myOctree = this->ss_myOctree;
	refined.push_back(temp);
	refined.back()->OS = this->OS;
}

//picks a uniformly distributed random point on the surface of the obj
Point Wavefront::Sample(float u1, float u2, Normal *Ns) const {
	Point sampledP;
	//first, pick a random face, then use u1, u2 for the individual face of the triangle
	float sampledArea = surfaceArea * RandomFloat();
	//do a binary search for the largest partial sum that is less that sampledArea
	int left = 0;
	int right = ntris - 1;
	int mid; //will contain the right index
	while(left < right) {
		mid = (int)((right-left)/2+left);
		if ((areaPartialSum[mid] < sampledArea) && (areaPartialSum[mid+1] > sampledArea)) {
			break;
		}
		else if ((areaPartialSum[mid-1] < sampledArea) && (areaPartialSum[mid] > sampledArea)) {
			mid--;
			break;
		}
		else if (sampledArea == areaPartialSum[mid])
			break;
		else if (sampledArea < areaPartialSum[mid])
			right = mid-1;
		else //if (sampledArea > areaPartialSum[mid])
			left = mid+1;
	}
	//from trianglemesh.cpp
	float b1, b2;
	UniformSampleTriangle(u1, u2, &b1, &b2);
	const Point &p1 = p[vertexIndex[3*mid]];
	const Point &p2 = p[vertexIndex[3*mid+1]];
	const Point &p3 = p[vertexIndex[3*mid+2]];
	sampledP = b1 * p1 + b2 * p2 + (1.f - b1 - b2) * p3;
	Normal n = Normal(Cross(p2-p1, p3-p1));
	*Ns = Normalize(n);
	if (reverseOrientation) *Ns *= -1.f;
	return sampledP; //ObjectToWorld(sampledP);
}

//for subsurface scattering
void Wavefront::GetSamples(int &nSamples, SamplePoint *&smpPts) {
	nSamples = npoints; //nverts;
	smpPts = new SamplePoint [npoints]; //[nverts];
	for(int i=0; i < npoints; i++) {
		smpPts[i].p = ObjectToWorld(p[i]);
		smpPts[i].area = vertexArea[i];
		smpPts[i].n = vertexNormal[i];
		//compute irradiance in photonmap?
		//smpPts[i].irradiance;
	}
	ss_n = nSamples;
	ss_smp = smpPts;
}

// CreateShape
// Only one parameter, filename, which directs the program
// to the relative path of the object file
extern "C" DLLEXPORT Shape *CreateShape(const Transform &o2w, bool reverseOrientation, const ParamSet &params) {		
		string filename = params.FindOneString("filename","");
		//used for subsurface scattering
		//float G = params.FindOneFloat("g", 0.79f);
		//float SigmaS = params.FindOneFloat("sigmas", 50.0f);
		//float SigmaA = params.FindOneFloat("sigmaa", 3.8f);
		return new Wavefront(filename.c_str(), o2w, reverseOrientation);
}

